#!/usr/bin/python
# -*- coding: utf-8 -*-
# Connect Domoticz to MQTT following the Homie Convention (https://github.com/homieiot/convention)

import json
import requests
import paho.mqtt.client as mqtt
import argparse
import base64

#import os
#import sys

#Global variables
lstdevices = []                     # Contains a list of all device objects
objMQTTClient = mqtt.Client()       # MQTTClient object
intDebug = 0                        # Debug level (0=no debug, 1=show logging information 2=lot of more info
strBaseTopic = "homie/domoticz"     # Base topic, under which all Domoticz devices are published.
strDomoticzURL = ""
strMQTTHost = ""


class Device:
############################################################################################################
# Class: Device
#-----------------------------------------------------------------------------------------------------------
# The device class holds a in memory representation of the device and it's status.
# The constructor is called, using the Domoticz JSON containing the device data. It's used to instantiate
# the object using the current status contained in the JSON. Furthermore it is used to determine the topics 
# to publish and subscribe to.
# Properties
# Name              - Name of the device
# DeviceId          - Unique identifier of the device, corresponds to Domoticz idx
# Type              - Type of device, determined by domoticz device type/subtype or combination
# lstProperties     - List of properties, determined by device type
# strProperties     - String containing the properties as published in MQTT $properties
# dictProperties    - Ditcionary containing the properties and their values.
############################################################################################################
    def __init__(self, jsonDeviceData):
        self.Name = jsonDeviceData["Name"]
        self.DeviceId = str((jsonDeviceData["idx"])).lower()    # The DeviceId is equal to the Domoticz idx
        self.Type = self._HomieMQTTType(jsonDeviceData)         # Standardize the device types
        self.lstProperties = self._HomieMQTTProperties()        # Populate the properties list (based on device type)
        self.strProperties = ""                                 # This will hold the properties to be published as needed for $properties
        self.dictProperties = {}                                # Dictionary containing property names and values
        
        if (int(jsonDeviceData["BatteryLevel"])<=100):          # If the device is battery operated, add the batterylevel property
            self.lstProperties.append("batterylevel");
        
        if (self.Type == "unknown"):                            # If the device type is not recognized, let it know
            print ("Unknown/Unsupported device type, not publishing: "+self.Name + " ["+self.DeviceId+"]")
        else:
            for property in self.lstProperties:                 # Here we populate the strProperties string based on found properties.
                if (self.strProperties == ""):                  # This is only used to publish the $properties topic
                    self.strProperties = self.strProperties + property
                else:
                    self.strProperties = self.strProperties + "," + property
            self.UpdateProperties(jsonDeviceData)               # Populate the properties with the actual data from Domoticz JSON
     

    def UpdateProperties(self, jsonDeviceData):
    ############################################################################################################
    # Class function: Device.UpdateProperties
    #-----------------------------------------------------------------------------------------------------------
    # Populates the dictProperties for properties in lstProperties with the actual values 
    # based in the supplied Domoticz JSON data.
    ############################################################################################################
        lstdatatypefloat = ["temperature","uv","visibility","direction","speed","gust","total","power","ozone","setpoint"]
        lstdatatypeinteger = ["humidity","barometer","level","maxlevel","illuminance"]
        lstdatatypestring = ["switch"]
        lstdatatypeboolean = ["dimmer"]
        lstdegreescelsius = ["temperature"]
        lstpercentage = ["humidity", "batterylevel"]
        lstnounit = ["switch","dimmer","level","maxlevel"]
        lstenum = ["selector"]
        lstsettable = ["switch","level","setpoint","selector"] # If it can be controlled, settable needs to be true. Also used for MQTT subscriptions.
        lstnotretained = [] # List of properties that should not be published as retained.
        
        for property in self.lstProperties:
            self.dictProperties[property+"/$name"] = property
            if (property in lstdatatypefloat):
                self.dictProperties[property+"/$datatype"] = "float"
            if (property in lstdatatypeinteger):
                self.dictProperties[property+"/$datatype"] = "integer"
            if (property in lstdatatypestring):
                self.dictProperties[property+"/$datatype"] = "string"
            if (property in lstdatatypeboolean):
                self.dictProperties[property+"/$datatype"] = "boolean"
            if (property in lstenum):
                self.dictProperties[property+"/$datatype"] = "enum"
                self.dictProperties[property+"/$format"] = self._enumselector(jsonDeviceData)
                
            
            #$settable
            if (property in lstsettable):
                self.dictProperties[property+"/$settable"] = "true"
            else:
                self.dictProperties[property+"/$settable"] = "false"
        
            #$retained
            if (property in lstnotretained):
                self.dictProperties[property+"/$retained"] = "false"
            else:
                self.dictProperties[property+"/$retained"] = "true"
        
            #$unit
            if (property in lstdegreescelsius):
                self.dictProperties[property+"/$unit"] = "°C"
            if (property in lstpercentage):
                self.dictProperties[property+"/$unit"] = "percentage"
            if (property in lstnounit):
                self.dictProperties[property+"/$datatype"] = ""
 
            # Populate dictProperties form Domoticz JSON data
            if (property == "temperature"):
                self.dictProperties[property] = str(jsonDeviceData["Temp"])
            if (property == "humidity"):
                self.dictProperties[property] = str(jsonDeviceData["Humidity"])
            if (property == "barometer"):
                self.dictProperties[property+"/$unit"] = "hPa"
                self.dictProperties[property] = str(jsonDeviceData["Barometer"])
            if (property == "uv"):
                self.dictProperties[property+"/$unit"] = "UVI"
                self.dictProperties[property] = str(jsonDeviceData["UVI"])
            if (property == "visibility"):
                self.dictProperties[property+"/$unit"] = "km"
                self.dictProperties[property] = str(jsonDeviceData["Visibility"])
            if (property == "direction"):
                self.dictProperties[property+"/$unit"] = "degrees"
                self.dictProperties[property] = str(jsonDeviceData["Direction"])
            if (property == "speed"):
                self.dictProperties[property+"/$unit"] = "m/s"
                self.dictProperties[property] = str(jsonDeviceData["Speed"])
            if (property == "gust"):
                self.dictProperties[property+"/$unit"] = "m/s"
                self.dictProperties[property] = str(jsonDeviceData["Gust"])
            if (property == "total"):
                self.dictProperties[property+"/$unit"] = "#"
                self.dictProperties[property] = str(jsonDeviceData["Counter"])
            if (property == "power"):
                self.dictProperties[property+"/$unit"] = "W"
                if ("Usage" in jsonDeviceData):
                    self.dictProperties[property] = str(jsonDeviceData["Usage"])[0:int(jsonDeviceData["Usage"].find(" "))]
                else:
                    self.dictProperties[property] = str(jsonDeviceData["Data"])[0:int(jsonDeviceData["Data"].find(" "))]
            if (property == "ozone"):
                self.dictProperties[property+"/$unit"] = "DU"
                self.dictProperties[property] = str(jsonDeviceData["Data"])[0:int(jsonDeviceData["Data"].find(" "))] 
            if (property == "switch"):
                if (str(jsonDeviceData["Status"]) == "Off"):
                    self.dictProperties[property] = "Off"
                else:
                    self.dictProperties[property] = "On"
            if (property == "dimmer"):
                if (str(jsonDeviceData["SwitchType"])=="Dimmer"):
                    self.dictProperties[property] = "true"
                else:
                    self.dictProperties[property] = "false"
            if (property == "maxlevel"):
                self.dictProperties[property] = str(jsonDeviceData["MaxDimLevel"])
            if (property == "level"):
                self.dictProperties[property] = str(jsonDeviceData["Level"])
            if (property == "illuminance"):
                self.dictProperties[property+"/$unit"] = "lux"
                self.dictProperties[property] = str(jsonDeviceData["Data"])[0:int(jsonDeviceData["Data"].find(" "))]
            if (property == "setpoint"):
                self.dictProperties[property] = str(jsonDeviceData["Data"])[0:int(jsonDeviceData["Data"].find(" "))]
            if (property == "batterylevel"):
                self.dictProperties[property] = str(jsonDeviceData["BatteryLevel"])
            if (property == "selector"):
                self.dictProperties[property] = self._LevelToEnum(jsonDeviceData["Level"])
                
    # Returns a comma seperated list of selector switch levels, comma's are removed, as are leading and trailing spaces.
    def _enumselector(self, jsonDeviceData):
    ############################################################################################################
    # Class function: Device._enumselector
    #-----------------------------------------------------------------------------------------------------------
    # Internal function to create HOMIE MQTT enum string based on Domoticz bas64 encoded level data.
    ############################################################################################################

        strLevelNames = str(base64.b64decode(jsonDeviceData["LevelNames"]), 'utf-8')
        strLevelNames = strLevelNames.replace(",","")  #remove comma's
        strLevelNames = strLevelNames.replace("|",",") #replace pipe with comma
        while (strLevelNames != strLevelNames.replace(" ,",",")): # Remove leading and trailing spaces in level names.
            strLevelNames = strLevelNames.replace(" ,",",")
        while (strLevelNames != strLevelNames.replace(", ",",")):
            strLevelNames = strLevelNames.replace(", ",",")
        strLevelNames = strLevelNames.lstrip()
        strLevelNames = strLevelNames.rstrip()
        return(strLevelNames)
    
    def _LevelToEnum(self, strLevel):
    ############################################################################################################
    # Class function: Device._LevelToEnum
    #-----------------------------------------------------------------------------------------------------------
    # Converts Domoticz level to corresponding enum value for selector switch
    ############################################################################################################
        intLevel = int(strLevel)
        intItteration = 0
        strFoundLevel = "-"
        strLevels = []
        intStartIndex = 0
        intNextItem = self.dictProperties["selector/$format"].find(",")
        while (intNextItem > 0):
            if (self.dictProperties["selector/$format"][intStartIndex:intNextItem] == strLevel):
                strFoundLevel = str(intItteration)
            if (intItteration == intLevel):
                strFoundLevel = self.dictProperties["selector/$format"][intStartIndex:intNextItem]
            intItteration = intItteration + 10
            intStartIndex = intNextItem + 1
            intNextItem = self.dictProperties["selector/$format"].find(",", intStartIndex)
        if (intItteration == intLevel):
                strFoundLevel = self.dictProperties["selector/$format"][intStartIndex:]
        return (strFoundLevel)
    
    # This function returns a list containing the relevant properties to be used for the device type.
    def _HomieMQTTProperties(self):
    ############################################################################################################
    # Class function: Device._HomieMQTTProperties
    #-----------------------------------------------------------------------------------------------------------
    # Determines MQTT HOMIE properties based on Device.Type
    ############################################################################################################
        properties = [self.Type]
        if (self.Type == "climate"):
            properties = ["temperature","humidity","barometer"]
        if (self.Type == "smartmeter"):
            properties = ["total","power"]
        if (self.Type == "wind"):
            properties = ["direction","speed","gust"]
        if (self.Type == "switch"):
            properties = ["switch","dimmer","maxlevel","level"]
        if (self.Type == "illuminance"):
            properties = ["illuminance"]
        if (self.Type == "thermostat"):
            properties = ["setpoint"]
        if (self.Type == "powerconsumption"):
            properties = ["power"]
        if (self.Type == "selector"):
            properties = ["level","selector"]
        return properties
    
    def _HomieMQTTType(self, deviceJSON): 
    ############################################################################################################
    # Class function: Device._HomieMQTTType
    #-----------------------------------------------------------------------------------------------------------
    # Determines MQTT HOMIE device type based on Domoticz Type and Subtypes
    ############################################################################################################
 
        lstSameName = ["Switch","Rain","UV","Wind","Thermostat"] # For these types the name in Domoticz and MQTT is the same.
        strReturnValue="unknown"

        if (deviceJSON["Type"] in lstSameName):
            strReturnValue = deviceJSON["Type"].lower() # Convert to lower case as per HOMIE convention.
        if (deviceJSON["Type"] == "Color Switch"): # Seting color is currently not supported, we'll handle it like a switch with dimmer function
            strReturnValue = "switch"
        if (deviceJSON["Type"] == "Lux"):
            strReturnValue = "illuminance"
        if (deviceJSON["Type"] == "Usage"):
            strReturnValue = "powerconsumption"
        if (deviceJSON["Type"] == "Temp"):
            strReturnValue = "temperature"
        if (deviceJSON["Type"] == "Temp + Humidity + Baro"):
            strReturnValue = "climate"
        if (deviceJSON["Type"] == "General"):
            if (deviceJSON["SubType"] == "Visibility"):
                strReturnValue = "visibility"
        if (deviceJSON["Type"] == "P1 Smart Meter"):
            strReturnValue = "smartmeter"
        if (deviceJSON["Type"] == "Lighting 1"):
            if (deviceJSON["SubType"] == "X10"):
                if (deviceJSON["SwitchType"] == "On/Off"):
                    strReturnValue = "switch"
                if (deviceJSON["SwitchType"] == "Push On Button"):
                    strReturnValue = "pushonbutton"
            if (deviceJSON["SwitchType"] == "Push Off Button"):
                strReturnValue = "pushoffbutton"
        if (deviceJSON["SubType"] == "Switch"):
            strReturnValue = "switch"
        if (deviceJSON["SubType"] == "Selector Switch"):
            strReturnValue = "selector"    
        if (deviceJSON["SubType"] == "kWh"):
            strReturnValue = "power"
        if (deviceJSON["SubType"] == "Custom Sensor"):
            if (deviceJSON["SensorUnit"] == "DU"):
                strReturnValue = "ozone"

        return strReturnValue
        
def DecodeDomoticzMessage(strMessage):
############################################################################################################
# Function: DecodeDomoticzMessage
#-----------------------------------------------------------------------------------------------------------
# Processes Domoticz/out JSON MQTT message
############################################################################################################
    global lstdevices
    
    domoticzMessage = json.loads(strMessage)

    for device in lstdevices:
        if device.DeviceId == str(domoticzMessage["idx"]):
            result = requests.get(strDomoticzURL+'/json.htm?type=devices&rid='+device.DeviceId)
            jsonDeviceData = json.loads(result.text)
            oldproperties = device.dictProperties.copy()
            device.UpdateProperties(jsonDeviceData["result"][0])
            for property, propertyvalue in device.dictProperties.items():
                if (propertyvalue != oldproperties[property]):  #Only publish if changed
                    MQTTPublish(strBaseTopic+"/"+device.DeviceId+"/"+property, propertyvalue, 0, True)
            break

def DomoticzCommand(strDeviceID, strDeviceProperty, strStatus):
############################################################################################################
# Function: DomoticzCommand
#-----------------------------------------------------------------------------------------------------------
# Sends Domoticz command to strDeviceID to change strDeviceProperty to strStatus
############################################################################################################
    global lstdevices
    
    strCommand = ""
    if (strDeviceProperty == "selector"):
        #Convert enum to level
        for device in lstdevices:
            if (device.DeviceId == strDeviceID):           
                if ((device.dictProperties["selector/$format"].find(strStatus)) > -1):
                    intLevel = (10 * (device.dictProperties["selector/$format"][:device.dictProperties["selector/$format"].find(strStatus)]).count(","))
                    strStatus = str(intLevel)
                    strDeviceProperty = "level"
                
    if (strDeviceProperty == "switch"):
        strCommand = strDomoticzURL+"/json.htm?type=command&param=switchlight&idx="+strDeviceID+"&switchcmd="+strStatus
    if (strDeviceProperty == "level"):
        strCommand = strDomoticzURL+"/json.htm?type=command&param=switchlight&idx="+strDeviceID+"&switchcmd=Set%20Level&level="+strStatus
    if (strDeviceProperty == "setpoint"):
        strCommand = strDomoticzURL+"/json.htm?type=command&param=setsetpoint&idx="+strDeviceID+"&setpoint="+strStatus
    
    if (strCommand != ""):
        if (intDebug > 0):
            print (strCommand)
        requests.get(strCommand)

        
        
def DecodeHomieMQTTMessage(strTopic, strPayload):
############################################################################################################
# Function: DecodeHomieMQTTMessage
#-----------------------------------------------------------------------------------------------------------
# Process incomming HOMIE MQTT message
############################################################################################################
    strBaseTopic = strTopic[:strTopic.rfind("/")]
    strDeviceProperty = strBaseTopic[strBaseTopic.rfind("/")+1:]
    if (intDebug > 0):
        print ("Property: "+strDeviceProperty)
    strDeviceTopic = strBaseTopic[:strBaseTopic.rfind("/")]
    strDeviceID = strDeviceTopic[strDeviceTopic.rfind("/")+1:]
    if (intDebug > 0):
        print ("DeviceID: "+strDeviceID)
    DomoticzCommand(strDeviceID, strDeviceProperty, strPayload)  # Send the set command to Domoticz
    
  
def on_connect(objMQTTClient, userdata, flags, rc):
############################################################################################################
# Function: on_connect
#-----------------------------------------------------------------------------------------------------------
# Callback function for MQTT connect
############################################################################################################
    if (intDebug > 0):
        print("MQTT: Connected with result code "+str(rc))

    # Subscribing in on_connect() means that if we lose the connection and
    # reconnect then subscriptions will be renewed.
    if (intDebug > 0):
        print ("MQTT: Subscribing to topic domoticz/out")
    objMQTTClient.subscribe("domoticz/out") # subscibe to domoticz/out
    for device in lstdevices:
        for property, propertyvalue in device.dictProperties.items():
            if (property[property.find("/")+1:]=="$settable"):
                # Subscibe to set for properties with settable = true
                if (propertyvalue == "true"):
                    if (intDebug > 0):
                        print("MQTT: Subscribing to topic "+strBaseTopic+"/"+device.DeviceId+"/"+property[:property.find("/")]+"/set")
                    result, mid = objMQTTClient.subscribe(strBaseTopic+"/"+device.DeviceId+"/"+property[:property.find("/")]+"/set")
  
def on_message(objMQTTClient, userdata, msg):
############################################################################################################
# Function: on_message
#-----------------------------------------------------------------------------------------------------------
# Callback function for MQTT when subscribed message is received
############################################################################################################
    if (intDebug > 1):
        print("MQTT received message -> Topic: "+msg.topic + " Payload: "+str(msg.payload))
        
    if ( msg.topic.find(strBaseTopic) == 0): # Determine if it's a Domoticz or Homie MQTT message
        DecodeHomieMQTTMessage(msg.topic, str(msg.payload,'utf-8'))
    else:
        DecodeDomoticzMessage(str(msg.payload, encoding='utf-8'))



def ParseArguments():
############################################################################################################
# Function: ParseArguments
#-----------------------------------------------------------------------------------------------------------
# Process command line arguments
############################################################################################################
    global intDebug  # Debug level
    global strDomoticzURL
    global strMQTTHost
    global strBaseTopic
    
    parser = argparse.ArgumentParser(description='Connect Domoticz to MQTT using the Homie Convention.')
    parser.add_argument("--debug", type=int, help="Log debug information to console.")
    parser.add_argument("--domoticz", help="Domoticz url (http://IP:PORT or http://FQDN:PORT)", required=True)
    parser.add_argument("--mqtt", help="MQTT host (IP or FQDN)", required=True)
    parser.add_argument("--basetopic", help="MQTT topic to use for publishing devices (default: homie/domoticz)")
    
    args = parser.parse_args()
    if args.debug:
        intDebug = args.debug
        print("Debug logging enabled.")
    strDomoticzURL = args.domoticz
    strMQTTHost = args.mqtt

    if (args.basetopic != None):
        strBaseTopic = args.basetopic


def MQTTPublish(strTopic, strPayload, intQos, blnRetain):
############################################################################################################
# Function: MQTTPublish
#-----------------------------------------------------------------------------------------------------------
# Publish MQTT topic
############################################################################################################
    if (intDebug > 0):
        if(strPayload != None):
            print ("MQTT publish: "+strTopic+"->"+strPayload)
        else:
            print ("MQTT publish: "+strTopic+"->None")
    objMQTTClient.publish(strTopic, strPayload, intQos, blnRetain)

############################################################################################################
# MAIN code
############################################################################################################

ParseArguments()

objMQTTClient.on_connect = on_connect
objMQTTClient.on_message = on_message

try:
    result = requests.get(strDomoticzURL+"/json.htm?type=devices&used=true")
except:
    print ("Could not connect to Domoticz. Check --domoticz command line argument.")
    quit()

domoticzresponse = json.loads(result.text)

# Now add all found devices to global variable lstdevices
for x in range(len(domoticzresponse["result"])):
  lstdevices.append(Device(domoticzresponse["result"][x]))

strNodes = ""
for device in lstdevices:
  if strNodes == "":
    strNodes = device.DeviceId
  else:
    strNodes = strNodes +","+ device.DeviceId
    
try:
    objMQTTClient.connect(strMQTTHost, 1883, 60)
except:
    print ("Could not connect to MQTT. Check --mqtt command line argument.")
    quit()

strHomieVersion = "4.0.0"

MQTTPublish(strBaseTopic+"/$state", "init", 0, True)
MQTTPublish(strBaseTopic+"/$homie", strHomieVersion, 0, True)
MQTTPublish(strBaseTopic+"/$name", "Domoticz", 0, True)
MQTTPublish(strBaseTopic+"/$nodes", strNodes, 0, True)
MQTTPublish(strBaseTopic+"/$extensions", None, 0, True)

for device in lstdevices:
    if (device.Type != "unknown"):
        strDeviceTopic = device.DeviceId
        MQTTPublish(strBaseTopic+"/"+strDeviceTopic+"/$name", device.Name, 0, True)
        MQTTPublish(strBaseTopic+"/"+strDeviceTopic+"/$type", device.Type, 0, True)
        MQTTPublish(strBaseTopic+"/"+strDeviceTopic+"/$properties", device.strProperties, 0, True)
        for x, y in device.dictProperties.items():   
            MQTTPublish(strBaseTopic+"/"+strDeviceTopic+"/"+x, y, 0, True)

MQTTPublish(strBaseTopic+"/$state", "ready", 0, True)

objMQTTClient.loop_forever()